﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Tempora
{
    internal class NtpPacket
    {

        public byte[] Bytes { get; private set; }

        private NtpPacket(byte[] bytes) {
            this.Bytes = bytes;
        }


        public static byte[] GetBytes(DateTime time, NtpPacket originPacket, DateTime arrivalTime) {
            var bytes = new List<byte>();

            var li = (int)NtpPacketLeapIndicator.None;
            var vn = (int)originPacket.VersionNumber;
            int mode;
            if (originPacket.Mode == NtpPacketMode.Client) {
                mode = (int)NtpPacketMode.Server;
            } else if (originPacket.Mode == NtpPacketMode.SymmetricActive) {
                mode = (int)NtpPacketMode.Server;
            } else {
                mode = (int)originPacket.Mode;
            }
            bytes.Add((byte)((li << 6) | (vn << 3) | (mode)));

            bytes.Add((byte)NtpPacketStratum.SecondaryReference3);
            bytes.Add(4); //PullInterval
            bytes.Add(0xfb); //Precision

            var rootDelay = BitConverter.GetBytes(1000);
            bytes.AddRange(new byte[] { rootDelay[3], rootDelay[2], rootDelay[1], rootDelay[0] });

            var rootDispersion = BitConverter.GetBytes(10000);
            bytes.AddRange(new byte[] { rootDispersion[3], rootDispersion[2], rootDispersion[1], rootDispersion[0] });

            bytes.AddRange(ASCIIEncoding.ASCII.GetBytes("LOCL"));

            bytes.AddRange(GetTimestamp(time.AddSeconds(-1)));
            bytes.AddRange(new byte[] { originPacket.Bytes[40], originPacket.Bytes[41], originPacket.Bytes[42], originPacket.Bytes[43], originPacket.Bytes[44], originPacket.Bytes[45], originPacket.Bytes[46], originPacket.Bytes[47] });
            bytes.AddRange(GetTimestamp(arrivalTime));
            bytes.AddRange(GetTimestamp(time));

            return bytes.ToArray();
        }


        public static NtpPacket ParseBytes(byte[] bytes) {
            var packet = new NtpPacket(bytes);

            packet.LeapIndicator = (NtpPacketLeapIndicator)(bytes[0] >> 6);
            packet.VersionNumber = (bytes[0] >> 3) & 0x07;
            packet.Mode = (NtpPacketMode)(bytes[0] & 0x07);

            packet.Stratum = (NtpPacketStratum)bytes[1];
            packet.PullInterval = new TimeSpan(0, 0, 0, (int)Math.Pow(2, bytes[2]));
            packet.Precision = new TimeSpan();

            packet.RootDelay = BitConverter.ToInt32(new byte[] { bytes[8], bytes[7], bytes[6], bytes[5] }, 0);

            packet.RootDispersion = BitConverter.ToInt32(new byte[] { bytes[12], bytes[11], bytes[10], bytes[9] }, 0); ;

            packet.ReferenceTimestamp = GetTimestamp(bytes, 16);
            packet.OriginateTimestamp = GetTimestamp(bytes, 24);
            packet.ReceiveTimestamp = GetTimestamp(bytes, 32);
            packet.TransmitTimestamp = GetTimestamp(bytes, 40);

            return packet;
        }

        /// <summary>
        ///  This is a warning of an impending leap second to be inserted in the NTP timescale. The bits are set before 23:59 on the day of insertion and reset after 00:00 on the following day. This causes the number of seconds (rollover interval) in the day of insertion to be increased or decreased by one. In the case of primary servers the bits are set by operator intervention, while in the case of secondary servers the bits are set by the protocol.
        /// </summary>
        public NtpPacketLeapIndicator LeapIndicator { get; private set; }

        /// <summary>
        /// NTP version. Must be 3 (as per RFC 1305).
        /// </summary>
        public int VersionNumber { get; private set; }

        /// <summary>
        /// This is a number indicating the association mode
        /// </summary>
        public NtpPacketMode Mode { get; private set; }

        /// <summary>
        /// This is a number indicating the stratum of the local clock
        /// </summary>
        public NtpPacketStratum Stratum { get; private set; }

        /// <summary>
        /// This is a number indicating the minimum interval between transmitted messages, in seconds as a power of two. For instance, a value of six indicates a minimum interval of 64 seconds.
        /// </summary>
        public TimeSpan PullInterval { get; private set; }

        /// <summary>
        /// This is a number indicating the precision of the various clocks, in seconds to the nearest power of two. The value must be rounded to the next larger power of two; for instance, a 50-Hz (20 ms) or 60-Hz (16.67 ms) power-frequency clock would be assigned the value -5 (31.25 ms), while a 1000-Hz (1 ms) crystal-controlled clock would be assigned the value -9 (1.95 ms).
        /// </summary>
        public TimeSpan Precision { get; private set; }

        /// <summary>
        /// This is a signed fixed-point number indicating the total roundtrip delay to the primary reference source at the root of the synchronization subnet, in seconds. Note that this variable can take on both positive and negative values, depending on clock precision and skew.
        /// </summary>
        public int RootDelay { get; private set; }

        /// <summary>
        /// This is a signed fixed-point number indicating the maximum error relative to the primary reference source at the root of the synchronization subnet, in seconds. Only positive values greater than zero are possible.
        /// </summary>
        public int RootDispersion { get; private set; }

        /// <summary>
        /// This is a 32-bit code identifying the particular reference clock. In the case of stratum 0 (unspecified) or stratum 1 (primary reference source), this is a four-octet, left-justified, zero-padded ASCII string. In the case of stratum 2 and greater (secondary reference) this is the four-octet Internet address of the primary reference host.
        /// </summary>
        public int ReferenceClockIdentifier { get; private set; }

        /// <summary>
        /// This field is the time the system clock was last set or corrected.
        /// </summary>
        public DateTime? ReferenceTimestamp { get; private set; }

        /// <summary>
        /// This is the time at which the request departed the client for the server.
        /// </summary>
        public DateTime? OriginateTimestamp { get; private set; }

        /// <summary>
        /// This is the time at which the request arrived at the server or the reply arrived at the client.
        /// </summary>
        public DateTime? ReceiveTimestamp { get; private set; }

        /// <summary>
        /// This is the time at which the request departed the client or the reply departed the server.
        /// </summary>
        public DateTime? TransmitTimestamp { get; private set; }



        private static DateTime? GetTimestamp(byte[] bytes, int offset) {
            long n1 = BitConverter.ToUInt32(new byte[] { bytes[offset + 3], bytes[offset + 2], bytes[offset + 1], bytes[offset + 0] }, 0);
            long n2 = BitConverter.ToUInt32(new byte[] { bytes[offset + 7], bytes[offset + 6], bytes[offset + 5], bytes[offset + 4] }, 0);

            if ((n1 == 0) && (n2 == 0)) { return null; }

            var time = new DateTime(1900, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);
            time = time.AddTicks(n1 * 10000000);
            time = time.AddTicks((long)(n2 / 4294967296.0 * 10000000));

            return time;
        }

        private static byte[] GetTimestamp(DateTime? timestamp) {
            if (timestamp == null) { return new byte[] { 0, 0, 0, 0, 0, 0, 0, 0 }; }

            DateTime time = timestamp.Value;
            if (time.Kind == DateTimeKind.Local) { time = time.ToUniversalTime(); }

            var eraTime = new DateTime(1900, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);
            var ts = time - eraTime;
            var n1 = (UInt32)ts.TotalSeconds;
            var buffer1LE = BitConverter.GetBytes(n1);

            var timeSec = eraTime.AddSeconds(n1);
            var n2 = (UInt32)Math.Min(Math.Max((time.Ticks - timeSec.Ticks) / 10000000.0 * 4294967296, 0), 4294967296);
            var buffer2LE = BitConverter.GetBytes(n2);

            return new byte[] { buffer1LE[3], buffer1LE[2], buffer1LE[1], buffer1LE[0], buffer2LE[3], buffer2LE[2], buffer2LE[1], buffer2LE[0] };
        }

    }


    /// <summary>
    /// This is a warning code of an impending leap second to be inserted/deleted in the last minute of the current day.
    /// </summary>
    internal enum NtpPacketLeapIndicator {
        /// <summary>
        /// No warning.
        /// </summary>
        None = 0,
        /// <summary>
        /// Last minute has 61 seconds.
        /// </summary>
        LastMinuteHas61Seconds = 1,
        /// <summary>
        /// Last minute has 59 seconds.
        /// </summary>
        LastMinuteHas59Seconds = 2,
        /// <summary>
        /// Alarm condition (clock not synchronized).
        /// </summary>
        ClockNotSynchronized = 3
    }



    /// <summary>
    /// This is a number indicating the protocol mode.
    /// </summary>
    internal enum NtpPacketMode {
        /// <summary>
        /// Unspecified.
        /// </summary>
        Unspecified = 0,
        /// <summary>
        /// Symmetric active.
        /// </summary>
        SymmetricActive = 1,
        /// <summary>
        /// Symmetric passive.
        /// </summary>
        SymmetricPassive = 2,
        /// <summary>
        /// Client.
        /// </summary>
        Client = 3,
        /// <summary>
        /// Server.
        /// </summary>
        Server = 4,
        /// <summary>
        /// Broadcast.
        /// </summary>
        Broadcast = 5,
        /// <summary>
        /// Reserved for NTP control messages.
        /// </summary>
        ReservedForNtpControlMessages = 6,
        /// <summary>
        /// Reserved for private use.
        /// </summary>
        ReservedForPrivateUse = 7
    }



    /// <summary>
    /// This is a number indicating the stratum.
    /// </summary>
    internal enum NtpPacketStratum {
        /// <summary>
        /// Kiss-o'-death message.
        /// </summary>
        Unknown = 0,
        /// <summary>
        /// Primary reference (e.g., calibrated atomic clock, radio clock).
        /// </summary>
        PrimaryReference = 1,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference1 = 2,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference2 = 3,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference3 = 4,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference4 = 5,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference5 = 6,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference6 = 7,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference7 = 8,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference8 = 9,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference9 = 10,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference10 = 11,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference11 = 12,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference12 = 13,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference13 = 14,
        /// <summary>
        /// Secondary reference (synchronized by NTP or SNTP).
        /// </summary>
        SecondaryReference14 = 15,
    }

}
